//-----------------------------------------------------------------------------
//
//	picdasm PIC microcontroller disassember
//
//-----------------------------------------------------------------------------
//
//	Cosmodog, Ltd.
//	415 West Superior Street
//	Chicago, IL   60610-3429
//	http://www.cosmodog.com
//
//-----------------------------------------------------------------------------
//
//	Copyright (C) 1999 Cosmodog, Ltd.
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//-----------------------------------------------------------------------------
//
//	This is the opcode definitions module.
//	As this is a disassembler it is not as obsessed with trapping every
//	invalid case (it assumes the incoming code is generally valid).
//
//-----------------------------------------------------------------------------

#include "includes.h"

//-----------------------------------------------------------------------------
//
// subSymbol -- return a pointer to a symbol if one exists
//

static const SYMBOLNAME *subSymbol(UINT8 value, const SYMBOLNAME *symList)
{
	int idx = 0;
	while(*symList[idx].name)
	{
		if(symList[idx].value == value)		// found a match, return it
		{
			break;
		}
		idx++;
	}
	return(&symList[idx]);				// return pointer to name or pointer to null if none
}


//-----------------------------------------------------------------------------
//
// specific opcode handlers
//

static BOOLEAN DasmLiteral(FILE *outFile, UINT16 instruction, UINT16 address, const PIC_DEF *procDef, UINT16 opNum)
{
	const OPCODE_DEF *opcode = &procDef->opcodeList[opNum];
	int k = instruction & opcode->k;				// find the target address

	fprintf(outFile,"0x%04x",k);
	return(TRUE);
}

//-----------------------------------------------------------------------------
//
static BOOLEAN DasmBit(FILE *outFile, UINT16 instruction, UINT16 address, const PIC_DEF *procDef, UINT16 opNum)
{
	const OPCODE_DEF *opcode = &procDef->opcodeList[opNum];
	int f = instruction & opcode->f;								// register address
	int b = (instruction & opcode->b)>>opcode->bposn;				// bit number
	const SYMBOLNAME *symbol = subSymbol(f,procDef->regFileList);	// try to get a symbol
	char fnum[] = "0x0000";
	char bnum[] = "0";												// bit number in ASCII
	const char *fdisp = fnum;
	const char *bdisp = bnum;

	sprintf(fnum,"0x%04x",f);										// format as number in case we have no symbol
	sprintf(bnum,"%01d",b);
	if(*symbol->name)												// if it points to a non-empty string,
	{
		fdisp = symbol->name;										// show the name
		if(symbol->bitNames)										// if that name has symbols for its bits
		{
			bdisp = symbol->bitNames[b];							// show the bit's symbol
		}
	}
	fprintf(outFile,"%s,%s",fdisp,bdisp);
	return(TRUE);
}


//-----------------------------------------------------------------------------
//
static BOOLEAN DasmByte(FILE *outFile, UINT16 instruction, UINT16 address, const PIC_DEF *procDef, UINT16 opNum)
{
	const OPCODE_DEF *opcode = &procDef->opcodeList[opNum];
	int f = instruction & opcode->f;	// register address
	int d = instruction & opcode->d;	// destination (0 = W, 1 = f)
	const SYMBOLNAME *symbol;

	if(opcode->f)						// if there is an opcode,
	{
		symbol = subSymbol(f,procDef->regFileList);		// substitute a symbol, if possible

		if(*symbol->name)								// if it points to a non-empty string,
		{
			fprintf(outFile,"%s",symbol->name);			// show the name
		}
		else
		{
			fprintf(outFile,"0x%04x",f);	// otherwise just show the address
		}

		if(opcode->d)						// if there is a destination,
		{
			if(!d)
			{
				fprintf(outFile,",W");		// add destination (working register)
			}
			else
			{
				fprintf(outFile,",F");		// add destination (back into the file register)
			}
		}
	}
	return(TRUE);
}


//-----------------------------------------------------------------------------

// PIC12C5XX instruction set
// 12-bit opcodes
//
#undef BYTEMASK
#undef BITMASK
#undef LITMASK
#undef GOTOMASK
#undef FMASK
#undef KMASK
#undef KMASKGOTO
#undef DMASK
#undef BMASK
#define BYTEMASK	0x0fc0		// opcode mask for byte-oriented instructions
#define BITMASK		0x0f00		// opcode mask for bit-oriented instructions
#define LITMASK		0x0f00		// opcode mask for literal and control operations except goto
#define GOTOMASK	0x0e00		// opcode mask for goto
#define FMASK		0x001f		// mask for f (file register #)
#define KMASK		0x00ff		// mask for k (constant), except goto
#define KMASKGOTO	0x01ff		// mask for k (constant) goto
#define DMASK		0x0020		// mask for d (destination), 0 = w, 1 = f
#define BMASK		0x00e0		// mask for b (bit field)
#define BSHIFT		5			// amount to shift the bit field down

static OPCODE_DEF PIC12C5XXopcodeList[] =
{
	// mnem.	opcode  mask			f		dest	b#		posn	k			handler for operand (if any)
	{"nop   ",	0x0000, 0x0fff,			0, 		0,		0,		0,		0x0000,		NULL},		// evaluated first so tris doesn't catch it falsely

	// literal and control ops
	{"andlw ",	0x0e00, LITMASK,		0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"call  ",	0x0900, LITMASK,		0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"clrwdt",	0x0004, 0x0fff,			0, 		0,		0,		0,		0,			NULL},
	{"goto  ",	0x0a00, GOTOMASK,		0, 		0,		0,		0,		KMASKGOTO,	DasmLiteral},
	{"iorlw ",	0x0d00, LITMASK,		0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"movlw ",	0x0c00, LITMASK,		0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"option",	0x0002, 0x0fff,			0, 		0,		0,		0,		0,			NULL},
	{"retlw ",	0x0800, LITMASK,		0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"sleep ",	0x0003, 0x0fff,			0, 		0,		0,		0,		0,			NULL},
	{"tris  ",	0x0000, 0x0ff8,			0x0007,	0,		0,		0,		0,			DasmByte},
	{"xorlw ",	0x0f00, LITMASK,		0,		0,		0,		0,		KMASK,		DasmLiteral},

	// byte-oriented file operations
	{"addwf ",	0x01c0, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"andwf ",	0x0140, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"clrf  ",	0x0060, BYTEMASK|DMASK,	FMASK,	0,		0,		0,		0, 			DasmByte},
	{"clrw  ",	0x0040, BYTEMASK|DMASK,	0,		0,		0,		0,		0, 			DasmByte},
	{"comf  ",	0x0240, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"decf  ",	0x00c0, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"decfsz",	0x02c0, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"incf  ",	0x0280, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"incfsz",	0x03c0, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"iorwf ",	0x0100, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"movf  ",	0x0200, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"movwf ",	0x0020, BYTEMASK|DMASK,	FMASK,	0,		0,		0,		0, 			DasmByte},
	{"rlf   ",	0x0340, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"rrf   ",	0x0300, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"subwf ",	0x0080, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"swapf ",	0x0380, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"xorwf ",	0x0180, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},

	// bit-oriented file operations
	{"bcf   ",	0x0400, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},
	{"bsf   ",	0x0500, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},
	{"btfsc ",	0x0600, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},
	{"btfss ",	0x0700, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},

	{"",0,0,0,0,0,0,0,NULL}
};


// PIC12C67X instruction set
// 14-bit opcodes
//
#undef BYTEMASK
#undef BITMASK
#undef LITMASK
#undef GOTOMASK
#undef FMASK
#undef KMASK
#undef KMASKGOTO
#undef DMASK
#undef BMASK
#undef BSHIFT
#define BYTEMASK	0x3f00		// opcode mask for byte-oriented instructions
#define BITMASK		0x3c00		// opcode mask for bit-oriented instructions
#define LITMASK		0x3f00		// opcode mask for literal and control operations except goto
#define GOTOMASK	0x3800		// opcode mask for goto and call
#define FMASK		0x007f		// mask for f (file register #)
#define KMASK		0x00ff		// mask for k (constant), except goto
#define KMASKGOTO	0x07ff		// mask for k (constant) goto
#define DMASK		0x0080		// mask for d (destination), 0 = w, 1 = f
#define BMASK		0x0380		// mask for b (bit field)
#define BSHIFT		7			// amount to shift the bit field down

static OPCODE_DEF PIC12C67XopcodeList[] =
{
	// mnem.	opcode  mask			f		dest	b#		posn	k			handler for operand (if any)
	{"nop   ",	0x0000, 0x0fff,			0, 		0,		0,		0,		0x0000,		NULL},		// evaluated first so tris doesn't catch it falsely

	// literal and control ops
	{"addlw ",	0x3e00, 0x3e00,			0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"andlw ",	0x3900, LITMASK,		0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"call  ",	0x3000, GOTOMASK,		0, 		0,		0,		0,		KMASKGOTO,	DasmLiteral},
	{"clrwdt",	0x0064, 0x0fff,			0, 		0,		0,		0,		0,			NULL},
	{"goto  ",	0x3800, GOTOMASK,		0, 		0,		0,		0,		KMASKGOTO,	DasmLiteral},
	{"iorlw ",	0x0d00, LITMASK,		0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"movlw ",	0x3000, 0x3c00,			0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"retfie ",	0x0009, 0x3fff,			0, 		0,		0,		0,		0,			NULL},
	{"retlw ",	0x3400, 0x3c00,			0, 		0,		0,		0,		KMASK,		DasmLiteral},
	{"sleep ",	0x0063, 0x3fff,			0, 		0,		0,		0,		0,			NULL},
	{"sublw  ",	0x3c00, 0x3e00,			0,		0,		0,		0,		KMASK,		DasmLiteral},
	{"xorlw ",	0x3a00, LITMASK,		0,		0,		0,		0,		KMASK,		DasmLiteral},

	// byte-oriented file operations
	{"addwf ",	0x0700, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"andwf ",	0x0500, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"clrf  ",	0x0180, BYTEMASK|DMASK,	FMASK,	0,		0,		0,		0, 			DasmByte},
	{"clrw  ",	0x0100, BYTEMASK|DMASK,	0,		0,		0,		0,		0, 			DasmByte},
	{"comf  ",	0x0900, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"decf  ",	0x0300, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"decfsz",	0x0b00, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"incf  ",	0x0a00, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"incfsz",	0x0f00, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"iorwf ",	0x0400, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"movf  ",	0x0800, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"movwf ",	0x0080, BYTEMASK|DMASK,	FMASK,	0,		0,		0,		0, 			DasmByte},
	{"rlf   ",	0x0d00, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"rrf   ",	0x0c00, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"subwf ",	0x0200, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"swapf ",	0x0e00, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},
	{"xorwf ",	0x0600, BYTEMASK,		FMASK,	DMASK,	0,		0,		0, 			DasmByte},

	// bit-oriented file operations
	{"bcf   ",	0x1000, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},
	{"bsf   ",	0x1400, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},
	{"btfsc ",	0x1800, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},
	{"btfss ",	0x1c00, BITMASK,		FMASK,	0,		BMASK,	BSHIFT,	0, 			DasmBit},

	{"",0,0,0,0,0,0,0,NULL}
};


// PIC16C5X instruction set
// 12-bit opcodes
//
static OPCODE_DEF PIC16C5XopcodeList[] =
{
	{"nop   ",	0x0000, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x0000, NULL},		// evaluated first so tris doesn't catch it falsely

	{"andlw ",	0x0e00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"call  ",	0x0900, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"clrwdt",	0x0004, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x0000, NULL},
	{"goto  ",	0x0a00, 0x0e00, 0x0000, 0x0000, 0x0000, 0, 0x01ff, DasmLiteral},
	{"iorlw ",	0x0d00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"movlw ",	0x0c00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"option",	0x0002, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x00ff, NULL},
	{"retlw ",	0x0800, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"sleep ",	0x0003, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x00ff, NULL},
	{"tris  ",	0x0000, 0x0ff8, 0x0007, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"xorlw ",	0x0f00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, NULL},

	{"addwf ",	0x01c0, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"andwf ",	0x0140, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"clrf  ",	0x0060, 0x0fe0, 0x001f, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"clrw  ",	0x0040, 0x0fe0, 0x0000, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"comf  ",	0x0240, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"decf  ",	0x00c0, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"decfsz",	0x02c0, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"incf  ",	0x0280, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"incfsz",	0x03c0, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"iorwf ",	0x0100, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"movf  ",	0x0200, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"movwf ",	0x0020, 0x0fe0, 0x001f, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"rlf   ",	0x0340, 0x0fc0, 0x001f, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"rrf   ",	0x0300, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"subwf ",	0x0080, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"swapf ",	0x0380, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"xorwf ",	0x0180, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},

	{"bcf   ",	0x0400, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},
	{"bsf   ",	0x0500, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},
	{"btfsc ",	0x0600, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},
	{"btfss ",	0x0700, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},

	{"",0,0,0,0,0,0,0,NULL}
};


// PIC17C4X instruction set
// 16-bit opcodes
static OPCODE_DEF PIC17C4XopcodeList[] =
{
//								f		d		b		bposn	k
	{"nop   ",	0x0000, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x0000, NULL},		// evaluated first so tris doesn't catch it falsely

	{"andlw ",	0x0e00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"call  ",	0x0900, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"clrwdt",	0x0004, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x0000, NULL},
	{"goto  ",	0x0a00, 0x0e00, 0x0000, 0x0000, 0x0000, 0, 0x01ff, DasmLiteral},
	{"iorlw ",	0x0d00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"movlw ",	0x0c00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"option",	0x0002, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x00ff, NULL},
	{"retlw ",	0x0800, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, DasmLiteral},
	{"sleep ",	0x0003, 0x0fff, 0x0000, 0x0000, 0x0000, 0, 0x00ff, NULL},
	{"tris  ",	0x0000, 0x0ff8, 0x0007, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"xorlw ",	0x0f00, 0x0f00, 0x0000, 0x0000, 0x0000, 0, 0x00ff, NULL},

	{"addwf ",	0x0e00, 0xfe00, 0x00ff, 0x0100, 0x0000, 0, 0x0000, DasmByte},
	{"addwfc",	0x1000, 0xfe00, 0x00ff, 0x0100, 0x0000, 0, 0x0000, DasmByte},
	{"andwf ",	0x2800, 0xfe00, 0x00ff, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"clrf  ",	0x0060, 0x0fe0, 0x001f, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"clrw  ",	0x0040, 0x0fe0, 0x0000, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"comf  ",	0x0240, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"decf  ",	0x00c0, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"decfsz",	0x02c0, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"incf  ",	0x0280, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"incfsz",	0x03c0, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"iorwf ",	0x0100, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"movf  ",	0x0200, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"movwf ",	0x0020, 0x0fe0, 0x001f, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"rlf   ",	0x0340, 0x0fc0, 0x001f, 0x0000, 0x0000, 0, 0x0000, DasmByte},
	{"rrf   ",	0x0300, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"subwf ",	0x0080, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"swapf ",	0x0380, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},
	{"xorwf ",	0x0180, 0x0fc0, 0x001f, 0x0020, 0x0000, 0, 0x0000, DasmByte},

	{"bcf   ",	0x0400, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},
	{"bsf   ",	0x0500, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},
	{"btfsc ",	0x0600, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},
	{"btfss ",	0x0700, 0x0f00, 0x001f, 0x0000, 0x00e0, 5, 0x0000, DasmBit},

	{"",0,0,0,0,0,0,0,NULL}
};

// labels for various bits
//
const static char *statusNames[] =       { "C", "DC", "Z", "PD", "TO", "PA0", "PA1", "PA2" };
const static char *statusNames12C5XX[] = { "C", "DC", "Z", "PD", "TO", "PA0", "-",   "GPWUF" };
const static char *statusNames12C67X[] = { "C", "DC", "Z", "PD", "TO", "RP0", "RP1", "IRP" };
const static char *statusNames16C505[] = { "C", "DC", "Z", "PD", "TO", "PA0", "-",   "RBWUF" };

const static char *portANames[] =   { "RA0", "RA1", "RA2", "RA3", "-",   "-",   "-",   "-" };
const static char *portBNames[] =   { "RB0", "RB1", "RB2", "RB3", "RB4", "RB5", "RB6", "RB7" };
const static char *portBNames6[] =  { "RB0", "RB1", "RB2", "RB3", "RB4", "RB5", "-",   "-" };
const static char *portCNames[] =   { "RC0", "RC1", "RC2", "RC3", "RC4", "RC5", "RC6", "RC7" };
const static char *portCNames6[] =  { "RC0", "RC1", "RC2", "RC3", "RC4", "RC5", "-",   "-" };
const static char *portGPNames[] =  { "GP0", "GP1", "GP2", "GP3", "GP4", "GP5", "-",   "-" };
const static char *portGPSNames[] = { "GP0", "GP1", "GP2", "GP3", "GP4", "GP5", "SDA",   "SCL" };

const static char *osccalNames6[] =  { "-", "-", "CAL0",   "CAL1",   "CAL2", "CAL3", "CAL4", "CAL5" };
const static char *osccalNames4a[] = { "-", "-", "CALSLW", "CALFST", "CAL0", "CAL1", "CAL2", "CAL3" };

const static char *intConNames[] = { "GPIF", "INTF", "TOIF", "GPIE", "INTE", "TOIE", "PEIE", "GIE" };

const static char *pir1Names[] = { "-", "-", "-", "-", "-", "-", "ADIF", "-" };
const static char *pie1Names[] = { "-", "-", "-", "-", "-", "-", "ADIE", "-" };
const static char *pconNames[] = { "-", "POR", "-", "-", "-", "-", "-", "-" };

const static char *adCon0Names[] = { "ADCS1", "ADCS0", "-",     "CHS1", "CHS0", "GO/DONE", "-", "ADON" };
const static char *adCon1Names[] = { "PCFG0", "PCFG1", "PCFG2", "-",    "-",    "-",       "-", "-" };

// register file map for 12C508/12C509/12CE518/12CE519

const static SYMBOLNAME PIC12C5XXregFileList[] =
{
	{"INDF",	0x00, NULL},
	{"TMR0",	0x01, NULL},
	{"PCL",		0x02, NULL},
	{"STATUS",	0x03, statusNames12C5XX},
	{"FSR",		0x04, NULL},
	{"OSCCAL",	0x05, osccalNames6},
	{"GPIO",	0x06, portGPSNames},
	{"",0,NULL},							// end of list
};

// register file map for 12C508A/12C509A

const static SYMBOLNAME PIC12C5XXAregFileList[] =
{
	{"INDF",	0x00, NULL},
	{"TMR0",	0x01, NULL},
	{"PCL",		0x02, NULL},
	{"STATUS",	0x03, statusNames12C5XX},
	{"FSR",		0x04, NULL},
	{"OSCCAL",	0x05, osccalNames6},
	{"GPIO",	0x06, portGPNames},
	{"",0,NULL},							// end of list
};



// register file map for 12C671/12C672
const static SYMBOLNAME PIC12C67XregFileList[] =
{
	{"INDF",	0x00, NULL},
	{"TMR0",	0x01, NULL},
	{"PCL",		0x02, NULL},
	{"STATUS",	0x03, statusNames12C67X},
	{"FSR",		0x04, NULL},
	{"GPIO",	0x05, portGPNames},
	{"PCLATH",	0x0a, NULL},
	{"INTCON",	0x0b, intConNames},
	{"PIR1",	0x0c, pir1Names},
	{"ADRES",	0x1e, NULL},
	{"ADCON0",	0x1f, adCon0Names},
	{"INDF",	0x80, NULL},
	{"OPTION",	0x81, NULL},
	{"PCL",		0x82, NULL},
	{"STATUS",	0x83, statusNames12C67X},
	{"FSR",		0x84, NULL},
	{"TRIS",	0x85, NULL},
	{"PCLATH",	0x8a, NULL},
	{"INTCON",	0x8b, intConNames},
	{"PIE1",	0x8c, pie1Names},
	{"PCON",	0x8e, pconNames},
	{"OSCCAL",	0x8f, osccalNames4a},
	{"ADCON1",	0x9f, adCon1Names},
	{"",0,NULL},							// end of list
};


// register file map for PIC16C52, PIC16C54s,
// PIC16CR54s, PIC16C55s, PIC16C56s, PIC16CR56s,
// PIC16C57s, PIC16CR57s
const static SYMBOLNAME PIC16C52regFileList[] =
{
	{"INDF",	0x00, NULL},
	{"TMR0",	0x01, NULL},
	{"PCL",		0x02, NULL},
	{"STATUS",	0x03, statusNames},
	{"FSR",		0x04, NULL},
	{"PORTA",	0x05, portANames},
	{"PORTB",	0x06, portBNames},
	{"PORTC",	0x07, portCNames},
	{"",0,NULL},							// end of list
};



// register file map for PIC16C58s, PIC16CR58s,
//
const static SYMBOLNAME PIC16C58regFileList[] =
{
	{"INDF",	0x00, NULL},
	{"TMR0",	0x01, NULL},
	{"PCL",		0x02, NULL},
	{"STATUS",	0x03, statusNames},
	{"FSR",		0x04, NULL},
	{"PORTA",	0x05, portANames},
	{"PORTB",	0x06, portBNames},
	{"",0,NULL},							// end of list
};


// register file map for 16C505/PIC16C58s/PIC16CR58s,
//
const static SYMBOLNAME PIC16C505regFileList[] =
{
	{"INDF",	0x00, NULL},
	{"TMR0",	0x01, NULL},
	{"PCL",		0x02, NULL},
	{"STATUS",	0x03, statusNames16C505},
	{"FSR",		0x04, NULL},
	{"OSCCAL",	0x05, osccalNames6},
	{"PORTB",	0x06, portBNames6},
	{"PORTC",	0x07, portCNames6},
	{"",0,NULL},							// end of list
};


// generic register file map
// for processors which don't have detailed
// support yet
//
const static SYMBOLNAME PICXXregFileList[] =
{
	{"",0,NULL},							// end of list
};


// processor definitions
//
const static PIC_DEF processorList[] =
{
	{"12C5XX",		PIC12C5XXopcodeList,	PICXXregFileList},			// generic 12c5xx
	{"16C5X",		PIC16C5XopcodeList,		PICXXregFileList},			// generic 16c5x
	{"17C4X",		PIC17C4XopcodeList,		PICXXregFileList},			// generic 17c4x

	{"12C508",		PIC12C5XXopcodeList,	PIC12C5XXregFileList},
	{"12C508A",		PIC12C5XXopcodeList,	PIC12C5XXAregFileList},
	{"12C509",		PIC12C5XXopcodeList,	PIC12C5XXregFileList},
	{"12C509A",		PIC12C5XXopcodeList,	PIC12C5XXAregFileList},


	{"12C671",		PIC12C67XopcodeList,	PIC12C67XregFileList},
	{"12C672",		PIC12C67XopcodeList,	PIC12C67XregFileList},
	{"12CE673",		PIC12C67XopcodeList,	PIC12C67XregFileList},
	{"12CE674",		PIC12C67XopcodeList,	PIC12C67XregFileList},

	{"12C5E518",	PIC12C5XXopcodeList,	PIC12C5XXregFileList},
	{"12C5E519",	PIC12C5XXopcodeList,	PIC12C5XXregFileList},

	{"16C52",		PIC16C5XopcodeList,		PIC16C52regFileList},
	{"16C54",		PIC16C5XopcodeList,		PIC16C52regFileList},
	{"16CR54",		PIC16C5XopcodeList,		PIC16C52regFileList},
	{"16C55",		PIC16C5XopcodeList,		PIC16C52regFileList},
	{"16C56",		PIC16C5XopcodeList,		PIC16C52regFileList},
	{"16CR56",		PIC16C5XopcodeList,		PIC16C52regFileList},
	{"16C57",		PIC16C5XopcodeList,		PIC16C52regFileList},
	{"16CR57",		PIC16C5XopcodeList,		PIC16C52regFileList},

	{"16C58",		PIC16C5XopcodeList,		PIC16C58regFileList},
	{"16CR58",		PIC16C5XopcodeList,		PIC16C58regFileList},

	{"16C505",		PIC16C5XopcodeList,		PIC16C505regFileList},

	{"", 0, 0}							// end of list
};


//-----------------------------------------------------------------------------


const PIC_DEF * InitDisassemble(char *partName)
{
	int idx = 0;
	int offset = 0;								// assume no "PIC" at the front
	const PIC_DEF *procDef = 0;

	while(partName[idx])						// ensure it's all upper case
	{
		partName[idx] = toupper(partName[idx]);
		idx++;
	}
	if(strncmp(partName,"PIC",3) == 0)			// if the user's argument has 'PIC' at the front,
	{
		offset = 3;								// skip the first three letters of the user's argument
	}

	idx = 0;
	while(*processorList[idx].name)
	{
		if(strcmp(processorList[idx].name,partName+offset) == 0)
		{
			procDef = &processorList[idx];
			break;
		}
		idx++;
	}
	return(procDef);
}

void DisassembleInstruction(FILE *outFile, UINT16 instruction, UINT16 address, const PIC_DEF *procDef)
{
	int idx = 0;

	fprintf(outFile,"0x%04x\t%04x\t",address/2,instruction);		// display the address and data (16 bit each)
	while(procDef->opcodeList[idx].mask)							// mask = zero indicates end of list (meaningless otherwise)
	{
		if(procDef->opcodeList[idx].opcode == (instruction & procDef->opcodeList[idx].mask) )
		{
			fprintf(outFile,"%s\t",procDef->opcodeList[idx].mnemonic);				// print the mnemonic
			if(procDef->opcodeList[idx].handler)
			{
				procDef->opcodeList[idx].handler(outFile,instruction, address, procDef, idx);		// handle the rest of it
			}
			fputc('\n',outFile);												// terminate the line
			break;
		}
		idx++;
	}
	if(!procDef->opcodeList[idx].mask)
	{
		fprintf(outFile,"unknown opcode\n");
	}
}

#define MAXNAMESLEN 70

void ShowDevices()
{
	int idx = 0;
	int length = 0;

	fprintf(stderr,"supported devices:\n");
	while(*processorList[idx].name)
	{
		length += fprintf(stderr,"  %s",processorList[idx].name);
		idx++;
		if(*processorList[idx].name)
		{
			fputc(',',stderr);			// comma if not the last one
			length++;
			if(length > MAXNAMESLEN)
			{
				fputc('\n',stderr);		// newline if line length exceeded
				length = 0;
			}
		}
		else
		{
			fputc('\n',stderr);			// newline if it is the  last one
		}
	}
}

